import Data.List (nub)
import Control.Monad (liftM, liftM2)
import Test.QuickCheck (quickCheck, quickCheckAll,
                        Arbitrary(arbitrary), oneof, sized)

type Name = String

-- warning, there is a potential conflict with Prop in QuickCheck
-- for that reason, it isn't included in the imports from QuickCheck
data Prop = Var Name
          | F
          | T
          | Not Prop
          | Prop :||: Prop
          | Prop :&&: Prop
          deriving Eq

showProp :: Prop -> String
showProp (Var x)    = x
showProp (F)        = "F"
showProp (T)        = "T"
showProp (Not p)    = par ("not " ++ showProp p)
showProp (p :||: q) = par (showProp p ++ " || " ++ showProp q)
showProp (p :&&: q) = par (showProp p ++ " && " ++ showProp q)

par :: String -> String
par s = "(" ++ s ++ ")"

type Valn = Name -> Bool

eval :: Valn -> Prop -> Bool
eval vn (Var x)    = vn x
eval vn (F)        = False
eval vn (T)        = True
eval vn (Not p)    = not (eval vn p)
eval vn (p :||: q) = eval vn p || eval vn q
eval vn (p :&&: q) = eval vn p && eval vn q

type Names = [Name]

names :: Prop -> Names
names (Var x)    = [x]
names (F)        = []
names (T)        = []
names (Not p)    = names p
names (p :||: q) = nub (names p ++ names q)
names (p :&&: q) = nub (names p ++ names q)


p0 :: Prop
p0 =  (Var "a" :&&: Var "b") :||:
      (Not (Var "a") :&&: Not (Var "b"))

p1 :: Prop
p1 =  (Var "a" :&&: Not (Var "a"))

vn0 :: Valn
vn0 "a" = False
vn0 "b" = False

test_Prop :: Bool
test_Prop =
  showProp p0 == "((a && b) || ((not a) && (not b)))" &&
  showProp p1 == "(a && (not a))" &&
  names p0 == ["a","b"]  &&
  eval vn0 p0 == True

empty :: Valn
empty = error "undefined"

extend :: Valn -> Name -> Bool -> Valn
extend vn x b y | x == y    = b
                | otherwise = vn y

valns :: Names -> [Valn]
valns []     = [ empty ]
valns (x:xs) = [ extend vn x b
                 | vn <- valns xs, b <- [True, False] ]

satisfiable :: Prop -> Bool
satisfiable p = or [ eval e p | e <- valns (names p) ]

test_satisfiable :: Bool
test_satisfiable = satisfiable p0 && not (satisfiable p1)

-- Simplify

isSimple :: Prop -> Bool
isSimple (Not p)    = isSimple p && not (isOp p)
  where
  isOp (Not p)      = True
  isOp (p :||: q)   = True
  isOp (p :&&: q)   = True
  isOp p            = False
isSimple (p :||: q) = isSimple p && isSimple q
isSimple (p :&&: q) = isSimple p && isSimple q
isSimple p          = True

simplify :: Prop -> Prop
simplify (Not p)    =  knot (simplify p)
  where
  knot (Not p)      = p
  knot (p :||: q)   = knot p :&&: knot q
  knot (p :&&: q)   = knot p :||: knot q
  knot p            = Not p
simplify (p :||: q) = simplify p :||: simplify q
simplify (p :&&: q) = simplify p :&&: simplify q
simplify p          = p

p2 :: Prop
p2 =  Not p0

test_simplify :: Bool
test_simplify =
  showProp p2 == "(not ((a && b) || ((not a) && (not b))))" &&
  showProp (simplify p2) == "(((not a) || (not b)) && (a || b))" &&
  not (isSimple p2) &&
  isSimple (simplify p2)

-- allow QuickCheck to generate arbitrary values of type Prop
instance Arbitrary Prop where
  arbitrary = sized prop
    where
    prop 0 =
      oneof [return F,
             return T,
             liftM Var arbitrary]
    prop n | n > 0 =
      oneof [return F,
             return T,
             liftM Var arbitrary,
             liftM Not (prop (n-1)),
             liftM2 (:&&:) (prop (n `div` 2)) (prop (n `div` 2)),
             liftM2 (:||:) (prop (n `div` 2)) (prop (n `div` 2))]
